# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2001,2002,2003,2004,2005 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: _envtoxml.awk,v 1.1 2004/08/08 21:15:38 carlo Exp $
# =====================================================================
# int _envtoxml([string template, string fileroot, string eval,\
#		string table, boolean trim, string tago, string tagc,\
#		string ofile, int lvl, boolean parse])
#
# Take environment variables and use them to replace special tags in
# an XML template file. All arguments are optional, but at least the
# 'template' parameter should be passed, otherwise the result will
# hardly be of any usefulness :-)
# =====================================================================

function _envtoxml(tpl,froot,eval,tbl,trim,tago,tagc,ofile,lvl,parse, \
	ltago,ltagc,retago,retagc,nl,end,tlen,tag,enc,a,i,\
	input,record,j,start) {

  i=j=0

  # Check args and set defaults.
  if (tpl == "") tpl = "/dev/null"
  if (ofile == "") ofile = _STDOUT
  if (froot == "") froot = ENVIRON["CSA_ROOT"] "/xml/"
  if (tago == "") tago = "$["
  if (tagc == "") tagc = "]"
  if (tago == tagc) return _EINVAL # start- and end-tags must be different
  if (!ENVIRON["CSA_TPLMAX"]) ENVIRON["CSA_TPLMAX"] = 10

  # prevent infinite recursions.
  if (lvl > ENVIRON["CSA_TPLMAX"]) return 1

  ltago = length(tago)
  ltagc = length(tagc)

  # Prepare the substitution regexp's.
  retago = _escreg(tago)
  retagc = _escreg(tagc)

  # tags can be expressed in several formats, including:
  #
  # 1) CSA standard (canonical) format:
  #
  #    $[tag:f]
  #
  #    where '$[' and ']' can be replaced with different sequences,
  #    and 'f' is the encoding flag, which is used only in external
  #    file inclusions.
  #
  # 2) XML format:
  #
  #    <csa:tag enc="f" />
  #
  #    Unfortunately the XML format is of very limited use, as it
  #    breaks the template well-formedness when used inside comment
  #    blocks, in attribute values and possibly in other places.
  #    A better syntax is definitely needed if I want tags to
  #    parse as well-formed XML in all circumstances, but so far
  #    I haven't been able to devise a general solution. At the
  #    moment I recommend that the canonical form be used.
  #    The XML form is experimental at best, and will definitely
  #    change in the future and may even be removed altogether, so
  #    don't use it.

  retago = retago "[-A-Za-z0-9_/.:]+([ \t]+enc *= *\".\")?" retagc

  if (!lvl) {				# only if 1st recursion.

     # Load an extra NoSQL table if requested. Table rows will be
     # assigned to _TBL[1], _TBL[2], etc., with _TBL[0] being the No. of
     # records in table. Any table header information is ignored. Note
     # that splitting table records into fields and unescaping NoSQL
     # special sequences is up to whoever will make use of these values,
     # often through the application-level _userproc() function.

     if (tbl != "") {
        _TBL[0] = 0
        while (getline record < tbl > 0) {
	   # Store table header in a separate place.
	   if (record ~ /^\001/) {
	      _TBLH = record
	      continue
	   }
	   # Append raw table record to _TBL[].
	   _TBL[++_TBL[0]] = record
        }

        close(tbl)
     }

     # This is mostly for 'csaExit.value'.

     if (ENVIRON["CSA_RPC_RESULT"] != _NULL) {
	# remove special value set by 'csaExit.value', if any.
	sub(/<csa:nil *\/>/,_NULL,ENVIRON["CSA_RPC_RESULT"])
	_OUT[0] = 1
	_OUT[1] = ENVIRON["CSA_RPC_RESULT"]
	gsub(/\t/,"\\t",_OUT[1])
	gsub(/\n/,"\\n",_OUT[1])
	gsub(/\001/,"\t",_OUT[1])

	_tbltordf(_OUTH,_OUT)

	# restore special pseudo-null value after calling _tbltordf().
	if (ENVIRON["CSA_RPC_RESULT"] == _NULL)
	   ENVIRON["CSA_RPC_RESULT"] = "<csa:nil />"

	# Mark the end of HTTP headers. Before this is done,
	# we may want to add further headers beside the basic
	# ones altready printed by 'csaTemplate'.

	print _NULL >> ofile
     }

     for (j in ENVIRON) {

         # Always skip rc(1) functions, confidential data
         # and other CSA-specific stuff.

         if (j ~ /(^(fn_|HTTP_COOKIE|CSA_AUTH_PW|csa0)|PGPPASS$)/) {
	    ENVIRON[j] = "<!-- removed -->"
	    continue
         }


         # If CSA_AUTH_SSL is in effect, let's be careful
         # about including HTTP_AUTHORIZATION in templates.

         if (j == "HTTP_AUTHORIZATION" && ENVIRON["CSA_AUTH_SSL"]) {
	      if (ENVIRON["SSL_CIPHER_USEKEYSIZE"] == "" && \
					ENVIRON["HTTPS"] != "on") {
	         # This cannot be set to anything other than "",
	         # or there will be problems with envtoxml.awk.
	         ENVIRON[j] = ""
	         continue
	      }
         }

	 # Only CSA_xxx variables are automatically encoded, while all
	 # others are left up to the application-level processor.

	 if (j ~ /^CSA_/) _response(j,ENVIRON[j])
     }

     # Test whether the write-back file already contains something.
     getline input < ENVIRON["CSA_TPLWB"]
     close(ENVIRON["CSA_TPLWB"])

     if (input != _NULL) {

	# if the write-back file is not empty it means that we are
	# already executing its content. To prevent csaTemplate()
	# from looping on '. $CSA_TPLWB' we must therefore clear the
	# file and make sure nothing else is written to it.

	_creat(ofile,_O_TRUNC)
	ENVIRON["CSA_TPLWB"] = "/dev/null"
     }

     # safety valve, to prevent the whole movie to loop forever in
     # case other instructions that cause further loops through
     # the template processing code are written to CSA_TPLWB
     # (like 'csaExit.ok' and similar things).

     else print "CSA_TPLWB=/dev/null" > ENVIRON["CSA_TPLWB"]

     # Call the application-level processor, unless 'csaExit.value'
     # was called (to return only one, selected value).

     if (ENVIRON["CSA_RPC_RESULT"] == _NULL) {
	_userproc(_O_RESPONSE)

	# handle template nullify requests issued by _sys(),
	# if called by _userproc() while processing responses.
	# Note that template nullification requests are taken into
	# consideration only if this is the first run.

	if (ENVIRON["CSA_TPLWB"] != "/dev/null" && \
		_RESULT == _ENOMSG) tpl = "/dev/null"
	else {
	   _tbltordf(_OUTH,_OUT)

	   # Mark the end of HTTP headers. Before this is done,
	   # we may want to add further headers beside the basic
	   # ones altready printed by 'csaTemplate'.

	   print _NULL >> ofile
	}
     }
  }

  if (eval != "") split(eval,a,",")

  # Process the XML template file.

  while (getline input < tpl > 0) {
     nl = 1
     do {
        match(input,retago)
        if (RSTART) {
	   end = index(input, tagc)
	   tlen = end - RSTART - ltago
	   start = RSTART		# save before calling _rdfprint()
	   _rdfprint(substr(input, 1, start - 1),parse,ofile)
	   tag = substr(input, start + ltago, tlen)

	   # Turn XML encoding attribute into canonical form.
	   sub(/[ \t]+enc *= *"/,":",tag); sub(/"/,_NULL,tag)

	   # Now handle encoding flags.
	   if (sub(/:c$/,_NULL,tag)) {
	      # character-data, to be included verbatim.
	      enc = "cdata"
	   }
	   else if (sub(/:p$/,_NULL,tag)) {
	      # parsed-character data.
	      enc = "pcdata"
	   }
	   else enc = "base64"			# default is binary.

	   for (i in a) {			# Is "tag" to be eval'ed ?
	       if (a[i] == tag) {
		  tag = ENVIRON[tag]		# Get tag value from env.
		  break
	       }
	   }

	   if (tag ~ /\//) {			# Catch file patterns.
	      if (tag ~ /\.\./) {		# Catch any '..' pattern.
		 _rdfprint(tago tag tagc,parse,ofile) # Print tag literally.
	      }
	      else {
		 # Read external file, if it exists.
		 if (froot != _NULL && tag !~ /^\//) {
		    tag = froot "/" tag
		    gsub(/\/\//, "/", tag)	# Turn any "//" into "/".
		 }
		 if (enc == "base64") {		# Base64.
		    if (ofile != _STDOUT) close(ofile)
		    system("exec csa-encdec -e -b " tag " >> " ofile)
		    if (ofile != _STDOUT) _creat(ofile,_O_APPEND)
		 }
		 else if (enc == "cdata") {
		    # embeddable character-data, include verbatim.
		    _envtoxml(tag,froot,eval,tbl,trim,tago,tagc,ofile,lvl+1,0)
		 }
		 else {
		    # parsed-character data, whatever "parsed" means in
		    # the specific context (XML in this case).
		    _envtoxml(tag,froot,eval,tbl,trim,tago,tagc,ofile,lvl+1,1)
		 }
		 nl = 0
	      }
	   }
	   else {
	      if (trim) ENVIRON[tag] = _strip(ENVIRON[tag])
	      _rdfprint(ENVIRON[tag],parse,ofile)
	   }

	   input = substr(input, end + ltagc)
	}
	else {
	   _rdfprint(input,parse,ofile)
	   input = _NULL		# Not to loop forever.
	}
     }  while (length(input))

     if (nl) _rdfprint("\n",0,ofile)
  }

  close(tpl)
}

