/* Written by and copyright Carlo Strozzi <carlos@linux.it>. */

/* encode: copy stdin to stdout, applying the requested escaping
   of special characters along the way. Of course this could be
   done just as well with a sed(1) script, but the latter would be
   much slower on large input, and could not cope easily with the
   transformation of newlines.

   Copyright (C) 2003 Carlo Strozzi <carlos@linux.it>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


   Usage:  encode [-p|-x]

   where '-p' stands for PYX and '-x' for XML.

   Notes:
   According to the XML specification, characters 0 through 31 and
   character 127 may not appare anywhere in adocument, not even
   as character references, such as &#31;. The only other possibility
   it to either remove them or turn them into something sensible.
   I took this second approach, whereby \001 (that is rc(1) list
   concatenation character) is replaced with a single blank. This is
   experimental, and may change in future versions of CSA.

 ****************************************************************************/

static char Psz_RCS[] = "$Date: 2003/03/17 11:37:29 $";

#include <stdio.h>
#include <stdlib.h>

static void usage() {
  fprintf(stderr,"usage: encode [-p|-x]\n");
}

int main(int argc, char **argv) {

   int c, enc='p', nl=0;

   if ((argc > 1) &&
       (!strcmp(argv[1],"-h") || !strcmp(argv[1],"--help"))) {
      usage();
      exit(1);
   }

   if (argc > 1) {
      if (!strcmp(argv[1],"-p")) enc='p';
      else if (!strcmp(argv[1],"-x")) enc='x';
      else {
      	 usage();
      	 exit(1);
      }
   }

   if (enc == 'p') {

      /* PYX escaping: \ \t \n */

      putchar('-');

      while ((c = getchar()) != EOF) {

   	    if (nl) {
	        printf("\\n");
	        nl=0;
	    }

	    switch (c) {
	      case '\n':
	        nl=1;
	        break;
	      case '\t':
	        printf("\\t");
	        break;
	      case '\001':
	        putchar(' ');
	        break;
	      case '\\':
	        printf("\\\\");
	        break;
	      default:
   	        putchar(c);
	    }
      }

      putchar('\n');
   }
   else {

      /* XML escaping: < " & 0x001 */

      while ((c = getchar()) != EOF) {

	    switch (c) {
	      case '<':
		printf("&lt;");
	        break;
	      case '"':
	        printf("&quot;");
	        break;
	      case '&':
		printf("&amp;");
	        break;
	      case '\001':
	        putchar(' ');
	        break;
	      default:
   	        putchar(c);
	    }
      }
   }

   exit(0);
}

/* End of program */
